#ifndef DFILEFORMAT_H_
#define DFILEFORMAT_H_

#include <map>
#include <vector>
#include <string>

typedef std::vector<bool> bit_vector;

class DFileFormat_writer
{
 public:
  DFileFormat_writer(const char *fname)
    {
      fp = fopen(fname, "wb");
      if(!fp)
	throw(std::string("unable to open file ") + std::string(fname));
    }

  DFileFormat_writer()
    {
      fp = 0;
    }

  ~DFileFormat_writer()
    {
      close();
    }

  void set_comment(const std::string &_comment_line)
    {
      comment_line = _comment_line;
    }

  void set_header_field(const std::string &_key, const std::string &_value)
    {
      header_fields[_key] = _value;
    }

  void set_header_field(const std::string &_key, int _value)
    {
      char temp[100];
      sprintf(temp, "%d", _value);
      set_header_field(_key, temp);
    }

  void set_header_field(const std::string &_key, float _value)
    {
      char temp[100];
      sprintf(temp, "%f", _value);
      set_header_field(_key, temp);
    }

  void write_header()
    {
      fprintf(fp, "# %s\n", comment_line.c_str());

      fprintf(fp, "# \n");

      std::map<std::string, std::string>::iterator fields_iterator = header_fields.begin();
      for( ; fields_iterator != header_fields.end(); ++fields_iterator)
	fprintf(fp, "# %s: %s\n", fields_iterator->first.c_str(), fields_iterator->second.c_str());

      fprintf(fp, "##\n");
    }

  template<class T>
    void write(const T &val)
    {
      fwrite(&val, sizeof(T), 1, fp);
    }

  template<class T>
    void write(const T *val, int count)
    {
      fwrite(val, sizeof(T), count, fp);
    }

  void write(const std::string &str)
    {
      write(str.c_str(), (int) str.size());
    }

  void close()
    {
      if(fp)
	fclose(fp);
      fp=0;
    }

  long long get_file_position()
    {
      return ftello(fp);
    }

  template<class T>
    void write_vector(const std::vector<T> &vec)
    {
      typename std::vector<T>::const_iterator iter = vec.begin();
      for( ; iter != vec.end(); ++iter)
	{
	  T val = *iter;
	  fwrite(&val, 1, sizeof(T), fp);
	}
    }

  void write_vector(const bit_vector &vec)
    {
      bit_vector::const_iterator iter = vec.begin();
      for( ; iter != vec.end(); ++iter)
	{
	  bool val = *iter;
	  fwrite(&val, 1, sizeof(val), fp);
	}
    }

  FILE *file_handle() { return fp; }

 protected:
  std::string comment_line;
  std::map<std::string, std::string> header_fields;
  FILE *fp;
};


class DFileFormat_reader
{
 public:
  DFileFormat_reader(const char *fname)
    {
      fp = fopen(fname, "rb");
      if(!fp)
	throw(std::string("unable to open file ") + std::string(fname));
    }

  ~DFileFormat_reader()
    {
      close();
    }

  std::string get_comment()
    {
      return comment_line;
    }

  std::string get_header_field_string(const std::string &key)
    {
      return header_fields[key];
    }

  int get_header_field_int(const std::string &key, int default_val = 0)
    {
      std::map<std::string, std::string>::iterator ptr = header_fields.find(key);

      if(ptr == header_fields.end())
	return default_val;
      else
	return atoi((ptr->second).c_str());
    }

  float get_header_field_float(const std::string &key)
    {
      return (float) atof(header_fields[key].c_str());
    }

  void read_header()
    {
      char temp[1024];

      fgets(temp, 1024, fp);
      comment_line = std::string(temp+2);

      while(fgets(temp, 1024, fp) && !(temp[0] == '#' && temp[1] == '#'))
	{
	  char *t=strchr(temp, ':');
	  if(!t)
	    continue;
	  
	  *t = 0;
	  std::string key(temp+2);
	  t++;
	  if(*t == ' ')
	    t++;
	  std::string value(t);

	  // remove newline
	  std::string::size_type ii = value.find("\n");
	  value.replace(ii, 1, "");

	  header_fields.insert(std::pair<std::string, std::string>(key, value));
	}
    }

  template<class T>
    void read(T &val)
    {
      fread(&val, sizeof(T), 1, fp);
    }

  template<class T>
    void read(T *val, int count)
    {
      fread(val, count, sizeof(T), fp);
    }

  template<class T>
    void read_vector(std::vector<T> &vec, int count)
    {
      vec = std::vector<T>(count);

      for(int i=0; i<count; i++)
	{
	  T val;
	  fread(&val, 1, sizeof(T), fp);
	  vec[i] = val;
	}
    }

  void read_vector(bit_vector &vec, int count)
    {
      vec = bit_vector(count);

      for(int i=0; i<count; i++)
	{
	  bool val;
	  fread(&val, 1, sizeof(bool), fp);
	  vec[i] = val;
	}
    }

  int read_int()
    {
      int val;
      read(val);
      return val;
    }

  void close()
    {
      if(fp)
	fclose(fp);
      fp=0;
    }

  bool is_eof()
    {
      return (feof(fp) == 0) ? false : true;
    }

  FILE *file_handle() { return fp; }

 protected:
  std::string comment_line;
  std::map<std::string, std::string> header_fields;
  FILE *fp;
};



#endif

