//
//  DLib: A simple image processing library.
//
//  David Crandall, 2003-2005
//  crandall@cs.cornell.edu
//
//  Please do not redistribute this code.
//
//
//
//
#include <DPlane.h>
#include <DGaussianModel.h>

template<class T>
class _DGaussianKernel : public _DPlane<T>
{
  public:
  _DGaussianKernel(const _DPlane<T> &other) : _DPlane<T>(other) {}

  _DGaussianKernel<T> &operator =(const _DPlane<T> &other)
    {
      *this = _DGaussianKernel(other);
      return *this;
    }


  // sigma is the standard deviation (NOT the variance)
  // width must be odd. If width is 0, the filter width is set
  // automatically to include 3 standard deviations of the
  // gaussian distribution.
  //
  _DGaussianKernel(double sigma, int width=0) : _DPlane<T>()
  {
    // if width is zero, set width to include 3-sigma of the
    // gaussian distribution
    if(width == 0)
      width = int(ceil(sigma * 3.0)) * 2 + 1;

    (*this) = _DPlane<T>(1,width);

    assert(width % 2 == 1);
    double norm = 1.0/(sigma * sqrt(2*M_PI));
    int half_width = width/2;
    for(int i=0; i<width; i++)
      (*this)[0][i] = T(norm * exp(-pow(double(i-half_width),2)/(2*sigma*sigma)));
  }

  // sigma is the standard deviation (NOT the variance)
  // width must be odd. If width is 0, the filter width is set
  // automatically to include 3 standard deviations of the
  // gaussian distribution.
  //
  _DGaussianKernel(double row_sigma, double col_sigma, int height=0, int width=0)
  {
    if(width == 0)
      width = int(ceil(row_sigma * 3.0)) * 2 + 1;
    if(height == 0)
      height = int(ceil(col_sigma * 3.0)) * 2 + 1;
    
    (*this) = _DPlane<T>(height, width);
    assert(width % 2 == 1 && height % 2 == 1);

    double norm = 1.0/(row_sigma * sqrt(2*M_PI)) * 1.0/(col_sigma * sqrt(2*M_PI));
    int half_width = width/2;
    int half_height = height/2;
    for(int i=0; i<height; i++)
      for(int j=0; j<width; j++)
        (*this)[i][j] = norm * exp( -((i-half_height)*(i-half_height)) / (2.0 * row_sigma * row_sigma)
				    -((j-half_width)*(j-half_width)) / (2.0 * col_sigma * col_sigma));
  }


  _DGaussianKernel(const _DMatrix<T> &sigma, float sigma_count=3) 
    {
      init(sigma, sigma_count);
    }

  _DGaussianKernel(const _DMatrix<T> &sigma, int height, int width) 
    {
      init(sigma, height, width);
    }

 protected:
  
  void init(const _DMatrix<T> &sigma, float sigma_count)
    {
      int width, height;
      
      width = int(ceil(sqrt(sigma[1][1]) * sigma_count) * 2 + 1);
      height = int(ceil(sqrt(sigma[0][0]) * sigma_count) * 2 + 1);	  

      init(sigma, height, width);
    }
  
  void init(const _DMatrix<T> &sigma, int height, int width)
    {
      _DMatrix<T> mean(2,1);
      mean[0][0] = height/2;
      mean[0][1] = width/2;
      (*this) = _DPlane<T>(exp(DGaussianModel<T>(sigma, mean).get_likelihood_plane(height, width)));
    }

}; 


typedef _DGaussianKernel<double> DGaussianKernel;
