#ifndef __DGAUSSIANSTATS_H__
#define __DGAUSSIANSTATS_H__

#include <DMatrix.h>
#include <vector>
#include <numeric>


// FIXME : This should do something useful if a singular covariance matrix is encountered.

// Throughout this file, "likelihood" means "log of the probability"

template<class T>
class DGaussianModel
{
 public:
  // construct an empty gaussian model
  //
  DGaussianModel();

  // learn a gaussian model from some data; observations in rows, dimensions in cols
  //
  DGaussianModel(const _DMatrix<T> &data);

  // build a gaussian model with the specified covariance and mean
  //
  DGaussianModel(const _DMatrix<T> &__covariance, const _DMatrix<T> &__mean, bool reverse_dims=false);

  // copy constructor
  //
  DGaussianModel(const DGaussianModel<T> &other);

  // learn a gaussian model from some data; observations in rows, dimensions in cols
  //
  void learn(const _DMatrix<T> &data);

  // compute the likelihood of some data, given the gaussian model.
  // observations are in rows, dimensions in columns.
  //
  T get_data_likelihood(const _DMatrix<T> &data);

  // compute likelihood assuming data is 1-D
  T get_data_likelihood(T data);

  _DMatrix<T> get_likelihood_pointwise(const _DMatrix<T> &data);

  // compute likelihood of data given model, but also output the likelihood of each observation individually 
  //  in the likelihoods vector.
  //
  T get_data_likelihood(const _DMatrix<T> &data, std::vector<T> &likelihoods);

  // compute the likelihoods of the data given the model, then remove the lowest-likelihood
  // trim_percent% of the data, and return the new likelihood. Upon return, result_likelihoods
  // and result_data contain the likelihoods of the remaining data and the remaining 
  // data matrix, respectively.
  //
  T get_data_likelihood_trimmed(const _DMatrix<T> &data, float trim_percent, std::vector<T> &result_likelihoods, _DMatrix<T> &result_data);

  // Compute a "likelihood map" where each pixel location stores the likelihood of the gaussian pdf at
  //  that point  
  // FIXME: presently only implemented for 2-D.
  //
  _DMatrix<T> get_likelihood_plane(int rows, int cols);
  _DMatrix<T> get_likelihood_plane(const DPoint &size) { return get_likelihood_plane(size.row(), size.col()); }

  
  DGaussianModel<T> &operator=(const DGaussianModel<T> &other);

  // returns log of scaling constant in front of Gaussian
  //
  T log_constant();

  // returns inverse covariance matrix
  //
  _DMatrix<T> inverse_covariance();

  // returns covariance matrix
  _DMatrix<T> covariance() const { return _covariance; }

  // return determinant of covariance
  T determinant_of_covariance() { compute_constants(); return _covariance.determinant(); }

  // return the mean, as a *column* vector
  //
  _DMatrix<T> mean() const { return _mean; }

  // return the mean, as a *point*
  // FIXME: rounds to ints, and only works for 2-D
  //
  DPoint mean_pt() const { return DPoint(int(_mean[0][0]), int(_mean[1][0])); }

  // change mean to new_mean
  //
  void set_mean(const _DMatrix<T> &new_mean);
  void set_mean(const DPoint &new_mean);
  void set_covariance(const _DMatrix<T> &new_covariance);
  void force_diagonal_covariance();

  // "rescale" model by a scale factor, adjusting the mean and covariance accordingly
  void rescale(float scale_factor);

  // return dimensionality of model
  //
  int dimensionality() const  { return _mean.rows(); }

  template<class T2>
    friend std::ostream & operator<<(std::ostream &os, const DGaussianModel<T2> &model);
  
  bool is_covariance_bad(double threshold = 1.0) const;

 protected:

  void compute_constants();
  void parameter_sanity_check();

 protected:
  _DMatrix<T> _covariance, _mean, _inverse_covariance;
  bool constants_clean;
  T log_c;

};



#endif
