//
//  DLib: A simple image processing library.
//
//  David Crandall, 2003-2005
//  crandall@cs.cornell.edu
//
//  Please do not redistribute this code.
//
//
//
//
#ifndef __IMAGE_H_
#define __IMAGE_H_

#include "DRect.h"
#include "DPlane.h"
#include "DIntegralImage.h"
#include <string.h>

template<class T>
class _DImage
{
  public:
  typedef enum { rgb24 } image_init_type;

  _DImage() 
    {
      plane_count=0;
      plane_data=0;
    }

  _DImage(int _planes, int _rows, int _cols, image_init_type type, T *buf)
    {
      if(type == rgb24)
	{
	  assert(_planes == 3);

	  plane_data = new _DPlane<T> *[_planes];
	  plane_count = _planes;
	  
	  for(int i=0; i<_planes; i++)
	    plane_data[i] = new _DPlane<T>(_rows, _cols);

	  const T *buf_cp = buf;
	  T *out1_ptr = plane_data[0]->data_ptr(), *out2_ptr = plane_data[1]->data_ptr(), *out3_ptr = plane_data[2]->data_ptr();
	  int sz = _rows*_cols;
	  for(int i = 0; i < sz; i++)
	    {
	      out1_ptr[i] = *(buf_cp++);
	      out2_ptr[i] = *(buf_cp++);
	      out3_ptr[i] = *(buf_cp++);
	    }

	}
      else
	assert(0);
    }
  
  _DImage(int _planes, int _rows, int _cols)
    {
      plane_data = new _DPlane<T> *[_planes];
      plane_count = _planes;
      
      for(int i=0; i<_planes; i++)
	plane_data[i] = new _DPlane<T>(_rows, _cols);
    }

    _DImage(int _planes)
    {
       plane_data = new _DPlane<T> *[_planes];
       plane_count = _planes;

       for(int i=0; i<_planes; i++)
          plane_data[i] = new _DPlane<T>;
    }

    ~_DImage()
      {
	if(plane_data)
	  {
	    for(int i=0; i<planes(); i++)
	      delete plane_data[i];
	    
	    delete[] plane_data;
	  }
      }

    _DImage(const _DImage &other)
      {
	plane_data = 0;
	plane_count=0;

	*this = other;
      }

    _DImage(const _DPlane<T> &other)
     {
       int _planes = 1;
       plane_data = new _DPlane<T> *[_planes];
       plane_count = _planes;

       plane_data[0] = new _DPlane<T>;
       (*plane_data[0]) = other;
     }

    _DPlane<T> &operator[](int plane) const
    {
       return *plane_data[plane];
    }

    _DImage &operator=(const double D)
      {
	for(int i=0; i<plane_count; i++)
	  (*this)[i] = T(D);

	return *this;
      }

    _DImage &operator=(const _DImage &other)
      {
       if(plane_data)
	 {
	   for(int i=0; i<plane_count; i++)
	     delete plane_data[i];

	   delete[] plane_data;
	 }

	plane_count = other.planes();
	plane_data = new _DPlane<T> *[other.planes()];

	for(int i=0; i<plane_count; i++)
	    plane_data[i] = new _DPlane<T>(other[i]);

	return *this;

      }

    // create the luma (grayscale) plane
    // We use the approximation 0.3 * (red plane) + 
    // 0.6 * (green plane) + 0.1 * (blue blane)
    //
    _DPlane<T> get_luma_plane() const
    {
      if(planes() == 1)
	return (*this)[0];
      else if(planes() == 3)
	{
	  _DPlane<T> luma(rows(), cols());

	  T *red_cp = (*this)[0][0], *green_cp = (*this)[1][0], *blue_cp = (*this)[2][0];
	  T *out_cp = luma[0];

	  int count = rows() * cols();
	  for(int i=0; i < count; i++)
	    {
	      out_cp[i] = T(0.3 * red_cp[i] + 0.6 * green_cp[i] + 0.1 * blue_cp[i]);
	    }
	  
	  return luma;
	}
      else
	throw std::string("cannot compute luma plane in DImage::get_luma_plane()");
    }

    _DImage<T> srgb_to_lab() const
      {
	const double Yn = 1.0, Xn = 0.98072, Zn = 1.1823;
	_DImage<T> result(3, rows(), cols());

	for(int i=0; i<rows(); i++)
	  for(int j=0; j<cols(); j++)
	    {
	      double R = (*this)[0][i][j], G = (*this)[1][i][j], B=(*this)[2][i][j];

	      const double a = 0.055;
	      R = (R <= 0.04045) ? (R/12.92) : pow((R + a)/(1 + a), 2.4);
	      G = (G <= 0.04045) ? (G/12.92) : pow((G + a)/(1 + a), 2.4);
	      B = (B <= 0.04045) ? (B/12.92) : pow((B + a)/(1 + a), 2.4);

	      double X, Y, Z;
	      X= 0.4124*R+ 0.3576*G+ 0.1805*B;
	      Y= 0.2126*R+ 0.7152*G+ 0.0722*B; 
	      Z= 0.0193*R+ 0.1192*G+ 0.9505*B;

	      double t_x = X / Xn, t_y = Y / Yn, t_z = Z / Zn;
	      const double thresh = pow(6/29.0, 3.0);
	      
	      double f_x = t_x > thresh ? pow(t_x, 1/3.0) : ((1.0/3) * pow((29/6.0),2.0) * t_x + 4.0/29);
	      double f_y = t_y > thresh ? pow(t_y, 1/3.0) : ((1.0/3) * pow((29/6.0),2.0) * t_y + 4.0/29);
	      double f_z = t_z > thresh ? pow(t_z, 1/3.0) : ((1.0/3) * pow((29/6.0),2.0) * t_z + 4.0/29);

	      result[0][i][j] = 116 * f_y - 16; 
	      result[1][i][j] = 500 * (f_x - f_y);
	      result[2][i][j] = 200 * (f_y - f_z);
	    }

	return result;
      }



    _DImage<T> bilinear_rescale(float row_factor, float col_factor)
      {
	_DImage<T> result(planes());

	for(int i=0; i<planes(); i++)
	  result[i] = (*this)[i].bilinear_rescale(row_factor, col_factor);
	
	return result;
      }

    _DImage<T> flip_horizontally() const
      {
	_DImage<T> result(planes());

	for(int k=0; k<planes(); k++)
	  result[k] = (*this)[k].flip_horizontally();
	  
	return result;
      }

    _DImage<T> extract(const DRect &rect) const
      {
	_DImage<T> result(planes());

	for(int i=0; i<planes(); i++)
	  result[i] = (*this)[i].extract(rect);

	return result;
      }

    int rows() const { return plane_data[0]->rows(); }
    int cols() const { return plane_data[0]->cols(); }
    int planes() const { return plane_count; }

    friend std::istream &operator>>(std::istream &is, _DImage<double> &matrix);

  protected:
    int plane_count;
    _DPlane<T> **plane_data;

};

typedef _DImage<double> DImage;

#endif
