//
//  DLib: A simple image processing library.
//
//  David Crandall, 2003-2005
//  crandall@cs.cornell.edu
//
//  Please do not redistribute this code.
//
//
//
//
#ifndef _DMATRIX_H_
#define _DMATRIX_H_

//#define GSL_SUPPORT
#ifdef GSL_SUPPORT
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#endif

#include <assert.h>
#include <iostream>
#include "DPoint.h"
#include "DRect.h"
#include <math.h>
#include <vector>
#include <complex>
#include <DMath.h>
#include <string.h>
#include <algorithm>

template<class T>
std::vector< std::pair<T, int> > unique_count(const std::vector<T> &vec)
{
  std::vector< std::pair<T,int> > result;

  if(vec.size() == 0)
    return result;

  typename std::vector<T>::const_iterator iter = vec.begin();
  T last = *iter;
  int last_count = 1;

  iter++;

  for( ; iter != vec.end(); ++iter)
    {
      if(*iter != last)
	{
	  result.push_back(std::pair<T, int> (last, last_count));
	  last = *iter;
	  last_count = 1;
	}
      else
	last_count++;
    }

  result.push_back(std::pair<T, int> (last, last_count));

  return result;
}

// This class is just a basic, templated 2-D array.
//
template<class T>
class _DTwoDimArray
{
 public:

  ///////////////////////////////////////////////////////
  // constructors
  _DTwoDimArray();
  _DTwoDimArray(int _rows, int _cols);
  _DTwoDimArray(int _rows, int _cols, const T *array);
  _DTwoDimArray(const _DTwoDimArray<T> &other);


  ///////////////////////////////////////////////////////
  // destructor
  ~_DTwoDimArray();


  ///////////////////////////////////////////////////////
  // assignment operator
  _DTwoDimArray<T> &operator=(const _DTwoDimArray<T> &other);


  ///////////////////////////////////////////////////////
  // element access
  inline T *operator[](int row) const { return data[row]; }


  ///////////////////////////////////////////////////////
  // informational 
  inline int rows() const { return _rows; }
  inline int cols() const { return _cols; }
  inline T *data_ptr() const { return data_area; }


 protected:

  void deallocate_storage();
  void initialize_storage();

  /// Array of pointers to the beginning of each matrix row in data_area
  T **data;

  /// Pointer to actual matrix data
  T *data_area;

  /// Size of matrix
  int _rows, _cols;
};




template<class T> 
class _DMatrix : public _DTwoDimArray<T>
{
 public:

  ///////////////////////////////////////////////////////
  // constructor
  _DMatrix() : _DTwoDimArray<T>() {}
  _DMatrix(int _rows, int _cols) : _DTwoDimArray<T>(_rows, _cols) {}
  _DMatrix(int _rows, int _cols, const T *array) : _DTwoDimArray<T>(_rows, _cols, array) {}
  _DMatrix(const std::vector<T> &vec);
  _DMatrix(int _rows, int _cols, const T val);

  typedef enum { random, identity, zeros, ones } matrix_init_type;
  _DMatrix(int _rows, int _cols, matrix_init_type type);


  ///////////////////////////////////////////////////////
  // assignment operator
  _DMatrix<T> &operator=(T val);


  ///////////////////////////////////////////////////////
  // element access
  inline T &operator[](const DPoint &pt) const;

  /// Extract and return a submatrix
  /// \param DRect specifying portion of matrix to extract
  /// \returns The specified submatrix
  _DMatrix<T> extract(const DRect &rect) const;

  /// Extract and return a single row of a matrix
  /// \param row index of row to return
  _DMatrix<T> extract_row(int row) const;
  std::vector<T> extract_row_as_vector(int row) const;

  /// Extract and return a single column of a matrix
  /// \param col index of col to return
  _DMatrix<T> extract_col(int col) const;

  /// Extract a set of rows or columns from a matrix
  _DMatrix<T> extract_rows(const std::vector<int> &row_list) const;
  _DMatrix<T> extract_cols(const std::vector<int> &col_list) const;

  /// Copy contents of another matrix into a portion of this matrix
  /// \param pt Specifies upper-left corner of destination submatrix
  /// \param source Matrix to be copied
  void set_submatrix(const DPoint &dest_pt, const _DMatrix<T> &source);
  void set_submatrix(const DRect &dest_pt, T val);

  void set_row(int row, const _DMatrix<T> &in);
  void set_col(int col, const _DMatrix<T> &in);

  void set_row(int row, const std::vector<T> &in);

  void copy_row_from(const _DMatrix<T> &other, int src_row, int dest_row);
  void copy_row_from(const DPoint &other, int dest_row);
  void copy_col_from(const _DMatrix<T> &other, int src_col, int dest_col);

  template<class T2>
    friend _DMatrix<T2> horiz_concat(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);
  template<class T2>
    friend _DMatrix<T2> vert_concat(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);


  ///////////////////////////////////////////////////////
  // informational

  /// Return size of matrix
  /// \returns size of matrix, encoded as a DPoint object
  DPoint size() const { return DPoint(rows(), cols()); }

  // returns true iff the given (i,j) location is valid given the matrix's size
  bool is_valid_index(const DPoint &pt) const
    {
      return(pt.row() >= 0 && pt.row() < rows() &&
	     pt.col() >= 0 && pt.col() < cols());
    }

  /// Returns true iff the two matrices have the same size
  template<class T2>
    friend bool same_size(const _DMatrix<T2> &matrix1, const _DMatrix<T2> &matrix2);

  /// Returns true iff at least one element of the matrix is undefined (nan)
  bool is_any_nan() const;
  bool is_any_inf() const;


  ///////////////////////////////////////////////////////
  // re-shaping

  /// Exchanges two rows of the matrix
  void swap_rows(int row1, int row2);

  /// Create a new matrix with the same data but different dimensions
  /// \param new_rows new number of rows
  /// \param new_cols new number of columns
  /// \returns A matrix of the specified size, filled in with elements from
  ///          the original matrix, in row-major order
  /// \note The number of elements must remain unchanged.
  _DMatrix<T> reshape(int new_rows, int new_cols);

  // "search and replace" : change every occurrence of src to dest
  //  operates *in place* (ip)
  //
  void search_and_replace_ip(T src, T dest);


  ///////////////////////////////////////////////////////
  // basic math operators

  _DMatrix<T> operator*(const _DMatrix<T> &other) const;
  _DMatrix<T> operator+(const _DMatrix<T> &other) const;
  _DMatrix<T> operator-(const _DMatrix<T> &other) const;
  _DMatrix<T> operator+(T value) const;
  _DMatrix<T> operator-(T value) const;
  _DMatrix<T> operator*(T value) const;
  _DMatrix<T> operator/(double value) const;

  _DMatrix<T> operator<(T value);
  _DMatrix<T> operator>(T value);
  _DMatrix<T> operator==(T value);

  template<class T2>
    friend _DMatrix<T2> operator+(T2 value, const _DMatrix<T2> &other);
  template<class T2>
    friend _DMatrix<T2> operator-(T2 value, const _DMatrix<T2> &other);
  template<class T2>
    friend _DMatrix<T2> operator*(T2 value, const _DMatrix<T2> &other);
  template<class T2>
    friend _DMatrix<T2> operator-(const _DMatrix<T2> &m);

  /// Multiply the contents of two m x n matrices element-by-element (like the .* operator of Matlab)
  /// \returns The m x n result of multiplication
  template<class T2>
    friend _DMatrix<T2> pointwise_multiply(const _DMatrix<T2> &matrix1, const _DMatrix<T2> &matrix2);

  /// Divide the contents of two m x n matrices element-by-element (like the ./ operator of Matlab)
  /// \returns The m x n result of division
  template<class T2>
    friend _DMatrix<T2> pointwise_divide(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);
  template<class T2>
    friend _DMatrix<T2> pointwise_min(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);
  template<class T2>
    friend _DMatrix<T2> pointwise_max(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);
  template<class T2>
    friend _DMatrix<T2> pointwise_min(const _DMatrix<T2> &m1, T2 val);
  template<class T2>
    friend _DMatrix<T2> pointwise_max(const _DMatrix<T2> &m1, T2 val);
  template<class T2> 
    friend _DMatrix<T2> operator<(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);
  template<class T2> 
    friend _DMatrix<T2> operator>(const _DMatrix<T2> &m1, const _DMatrix<T2> &m2);

  // in-place operators
  void operator+=(const _DMatrix<T> &other);
  void operator-=(const _DMatrix<T> &other);
  void operator+=(T value);
  void operator-=(T value);
  void operator*=(double value);
  void operator/=(T value);


  ///////////////////////////////////////////////////////
  // more advanced math

  /// Compute absolute value of each matrix element
  /// \param matrix an m x n matrix
  /// \returns an m x n matrix containing the absolute values
  template<class T2>
    friend _DMatrix<T2> fabs(const _DMatrix<T2> &m);

  template<class T2>
    friend _DMatrix<T2> exp(const _DMatrix<T2> &m, bool fast);
  template<class T2>
    friend _DMatrix<T2> exp(const _DMatrix<T2> &m);

  template<class T2>
    friend _DMatrix<T2> log(const _DMatrix<T2> &m);

  /// Compute square root of each matrix element
  /// \param matrix an m x n matrix
  /// \returns an m x n matrix containing the square roots
  template<class T2>
    friend _DMatrix<T2> sqrt(const _DMatrix<T2> &matrix);

  /// Compute square of each matrix element
  /// \param matrix an m x n matrix
  /// \returns an m x n matrix containing the squares
  template<class T2>
    friend _DMatrix<T2> sqr(const _DMatrix<T2> &matrix);

  //#ifdef GSL_SUPPORT
  /// Computes the matrix inverse
  /// \returns The inverse (a matrix)
  _DMatrix<T> inverse() const;

  /// Computes the matrix determinant
  /// \returns The determinant (a scalar)
  T determinant() const;
  _DMatrix<T> eigen(_DMatrix<T> &vectors) const;

  void svd(_DMatrix<T> &u, _DMatrix<T> &s, _DMatrix<T> &v) const;

#ifdef GSL_SUPPORT
  template<class T2>
    friend gsl_matrix *DMatrix_to_gsl(const _DMatrix<T2> &dm);
  template<class T2>
    friend _DMatrix<T2> gsl_to_DMatrix(gsl_matrix *gm);
  template<class T2>
    friend _DMatrix<T2> gsl_to_DMatrix(gsl_vector *gm);
#endif

// in-place sqr() 
  void sqr_ip(void);

  /// Compute matrix transpose
  /// \returns transpose of matrix
  _DMatrix<T> transpose() const;

  _DMatrix<T> rotate_180() const;
  _DMatrix<T> flip_horizontal() const;
  _DMatrix<T> flip_vertical() const;


  _DMatrix<T> LU_factor(void);


  bool is_diagonal() const
    {
      for(int i=0; i<rows(); i++)
	for(int j=0; j<cols(); j++)
	  if(i != j && (*this)[i][j] != 0)
	    return false;

      return true;
    }

  std::vector<DPoint> sample_probabilities(int sample_count = 1) const;
  std::vector<DPoint> sample_likelihoods(int sample_count = 1, bool fast=false) const;




  ///////////////////////////////////////////////////////
  // statistical stuff

  /// Compute the median of all matrix elements
  /// \returns the median (a scalar)
  T median() const;

  /// Compute the mean of all matrix elements
  /// \returns the mean (a scalar)
  T mean() const;

  /// Computes the sum of all matrix elements
  /// \returns the sum (a scalar)
  T sum() const;

  /// Computes the sum of each matrix row
  /// \returns an mx1 matrix with one element for each row 
  _DMatrix<T> sum_rows() const;

  /// Computes the sum of each matrix row
  /// \returns a 1xm matrix with one element for each column
  _DMatrix<T> sum_cols() const;

  /// Computes the covariance matrix for the observations in the rows of this matrix.
  /// That is, assumes that this matrix is n x p, where n is the number of observations
  /// and p is the number of dimensions of each observation.
  /// \returns The p x p covariance matrix.
  _DMatrix<T> covariance(void) const;

  /// Computes the covariance matrix for the observations in the rows of this matrix.
  /// \param mean_vector The mean vector for the observations.
  /// \returns The p x p covariance matrix.
  _DMatrix<T> covariance(const _DMatrix<T> &mean_vector) const;

  /// Computes the mean of each matrix column.
  /// \returns A 1 x n (where original matrix was m x n) matrix containing the mean
  /// of each column.
  _DMatrix<T> means(void) const;

  _DMatrix<T> trimmed_means(int leaveout_count) const;

  /// Finds the maximum element in the matrix
  /// \returns The maximum element (a scalar)
  T max() const
    {
      int out_row, out_col;
      return max(out_row, out_col);
    }

  /// Finds the maximum element in the matrix, also returning the position of the element
  /// \param[out] out_row row of maximum element
  /// \param[out] out_col column of maximum element
  /// \returns The maximum element (a scalar)
  T max(int &out_row, int &out_col) const;
  T max(DPoint &out_location) const
    {
      int row, col;
      T val = max(row, col);

      out_location = DPoint(row, col);
      return val;
    }

  // Returns the same value as max(). If there is a unique global maximum,
  // also returns the same point as max(); otherwise returns a randomly-chosen
  // maximum.
  T max_random(DPoint &out_location) const;

  /// Finds the minimum element in the matrix
  /// \returns The minimum element (a scalar)
  T min() const
    {
      int out_row, out_col;
      return min(out_row, out_col);
    }

  /// Finds the minimum element in the matrix, also returning the position of the element
  /// \param[out] out_row row of minimum element
  /// \param[out] out_col column of minimum element
  /// \returns The minimum element (a scalar)
  T min(int &out_row, int &out_col) const;
  T min(DPoint &out_location) const
    {
      int row, col;
      T val = min(row, col);

      out_location = DPoint(row, col);
      return val;
    }


  ///////////////////////////////////////////////////////
  // sorting

  void sort_rows();
  void sort_cols();
  void sort_wholerows();
  _DMatrix<T> unique_rows(std::vector<int> &row_counts) const;


  ///////////////////////////////////////////////////////
  // i/o routines

  template<class T2>
    friend std::istream &operator>>(std::istream &is, _DMatrix<T2> &matrix);
  template<class T2>
    friend std::ostream &operator<<(std::ostream &is, const _DMatrix<T2> &matrix);	
  template<class T2>
    friend void fread(_DMatrix<T2> &matrix, FILE *fp, bool enforce_size);
  template<class T2>
    friend void fread(_DMatrix<T2> &matrix, FILE *fp);
  template<class T2>
    friend void fwrite(const _DMatrix<T2> &matrix, FILE *fp);

  ///////////////////////////////////////////////////////
  // changing types

  template<class T2, class T3>
    friend  void change_type(const _DMatrix<T2> &in, _DMatrix<T3> &out);


  using _DTwoDimArray<T>::rows;
  using _DTwoDimArray<T>::cols;
  using _DTwoDimArray<T>::data_ptr;
  using _DTwoDimArray<T>::operator[];

 protected:
  using _DTwoDimArray<T>::_rows;
  using _DTwoDimArray<T>::_cols;
  using _DTwoDimArray<T>::data;
  using _DTwoDimArray<T>::data_area;
  using _DTwoDimArray<T>::initialize_storage;

};







typedef _DMatrix<double> DMatrix;

static std::string basename(const std::string &str)
{
  char str2[str.size()+1];
  strcpy(str2, str.c_str());

  return(std::string(basename(str2)));
}

#endif
