//
//  DLib: A simple image processing library.
//
//  David Crandall, 2003-2005
//  crandall@cs.cornell.edu
//
//  Please do not redistribute this code.
//
//
//
//
#ifndef __DMULTIDMATRIX_H_
#define __DMULTIDMATRIX_H_

#include "DRect.h"
#include "DMatrix.h"



////////////////////////////////////////////// lazy version

template<class T>
class _DLazyMultiDMatrix
{
 public:
  _DLazyMultiDMatrix() 
    {
      mat_count = 0;
    }

  _DLazyMultiDMatrix(int _dim_count, int p1) : mat_count(p1)
    {
      //      assert(_dim_count == 3);
    }
  
  virtual _DMatrix<T> &get(int c1) const = 0;

  int planes() const { return mat_count; }

  virtual ~_DLazyMultiDMatrix() { }

 protected:
  int mat_count;
};



template<class T>
class _DMultiDMatrix : public _DLazyMultiDMatrix<T>
{
 public:
  _DMultiDMatrix() 
    {
      dimensions = 0;
      dim_sizes = 0;
      matrices = 0;
      //      mat_count=0;
    }

  _DMultiDMatrix(int _dim_count, int p1, int p2, int p3) 
    {
      assert(_dim_count == 3);
	
      int dims[3];
      dims[0] = p1;
      dims[1] = p2;
      dims[2] = p3;

      allocate_storage(3, dims);
    }
  
  _DMultiDMatrix(int _dim_count, int p1, int p2, int p3, int p4)
    {
      assert(_dim_count == 4);
	
      int dims[4];
      dims[0] = p1;
      dims[1] = p2;
      dims[2] = p3;
      dims[3] = p4;

      allocate_storage(4, dims);
    }

  _DMultiDMatrix(int _dim_count, int *_dims)
    {
      assert(_dim_count > 2);

      allocate_storage(_dim_count, _dims);
    }
  

  void allocate_storage(int _dim_count, int *_dims)
    {
      dimensions = _dim_count;
      dim_sizes = new int[dimensions];
      for(int i=0; i<dimensions; i++)
	dim_sizes[i] = _dims[i];

      mat_count = 1;
      for(int i=0; i<dimensions-2; i++)
	mat_count *= _dims[i];

      matrices = new _DMatrix<T> *[mat_count];

      for(int i=0; i<mat_count; i++)
	matrices[i] = new _DMatrix<T>(_dims[dimensions-2], _dims[dimensions-1]);    
    }

  
  virtual ~_DMultiDMatrix()
    {
      if(dim_sizes)
	delete[] dim_sizes;
      
      if(matrices)
	{
	  for(int i=0; i<mat_count; i++)
	    delete matrices[i];
	  
	  delete[] matrices;
	}
    }
  
  _DMultiDMatrix(const _DMultiDMatrix<T> &other)
    {
      matrices = 0;
      dim_sizes = 0;

      *this = other;
    }

  _DMultiDMatrix<T> operator*(const T &D) const
    {
      _DMultiDMatrix<T> result(*this);

      for(int i=0; i<mat_count; i++)
	result.get(i) *= D;

      return result;
    }


  _DMultiDMatrix<T> &operator=(const T &D) 
    {
      for(int i=0; i<mat_count; i++)
	get(i) = D;

      return *this;
    }

  _DMultiDMatrix<T> &operator=(const _DMultiDMatrix<T> &other)
    {
      if(&other == this)
	return *this;

      if(matrices)
	{
	  for(int i=0; i<mat_count; i++)
	    delete matrices[i];

	  delete[] matrices;
	}

      if(dim_sizes)
	delete[] dim_sizes;

      mat_count = other.planes();
      matrices = new _DMatrix<T> *[mat_count];
      dimensions = other.get_dimensions_count();

      for(int i=0; i<mat_count; i++)
	matrices[i] = new _DMatrix<T>(other.get(i));

      dim_sizes = new int[dimensions];
      for(int i=0; i<dimensions; i++)
	dim_sizes[i] = other.get_dims(i);

      return *this;
    }

  int get_dimensions_count() const
    {
      return dimensions;
    }

  _DMatrix<T> &get(int c1) const
    {
      return *(matrices[c1]);
    }

  _DMatrix<T> &get(int c1, int c2) const
    {
      return *(matrices[c1*dim_sizes[1]+c2]);
    }

  _DMultiDMatrix<T> pointwise_min(const _DMultiDMatrix<T> &m1) const;
  _DMultiDMatrix<T> pointwise_max(const _DMultiDMatrix<T> &m1) const;

  template<class T2>
    friend _DMatrix<T2> pointwise_min(const _DMultiDMatrix<T2> &m1, _DMatrix<int> &index);
  template<class T2>
    friend _DMatrix<T2> pointwise_max(const _DMultiDMatrix<T2> &m1, _DMatrix<int> &index);


  int get_dims(int i) const
    {
      if(!dim_sizes)
	return 0;
      else
	return dim_sizes[i];
    }

  template<class T2>
    friend bool same_size(const _DMultiDMatrix<T2> &m1, const _DMultiDMatrix<T2> &m2);


  int rows() const 
    { 
      return matrices[0]->rows();
    }
  int cols() const 
    { 
      return matrices[0]->cols();
    }

  DPoint size() const
    {
      return DPoint(rows(), cols());
    }
  //  int planes() const { return mat_count; }

  friend std::istream &operator>>(std::istream &is, _DMultiDMatrix<T> &matrix)
    {
      int dimensions;
      is >> dimensions;

      if(dimensions)
	{
	  int  *dims = new int[ dimensions ];

	  memset (dims, 0, dimensions * sizeof( int ));

	  for(int i=0; i<dimensions; i++)
	    is >> dims[i];
	  
	  matrix = _DMultiDMatrix<T>(dimensions, dims);
	  
	  for(int i=0; i<matrix.planes(); i++)
	    is >> matrix.get(i);

	  delete [] dims;
	}
      return is;
    }

  friend void fread(_DMultiDMatrix<T> &matrix, FILE *fp)
    {
      int dimensions;
      fread(&dimensions, 1, sizeof(int), fp);

      if(dimensions)
	{
	  int dims[dimensions];
	  for(int i=0; i<dimensions; i++)
	    fread(dims+i, 1, sizeof(int), fp);

	  matrix = _DMultiDMatrix<T>(dimensions, dims);

	  for(int i=0; i<matrix.planes(); i++)
	    fread(matrix.get(i), fp);
	}
    }


  friend std::ostream &operator<<(std::ostream &os, const _DMultiDMatrix<T> &matrix)
    {
      os << matrix.dimensions << std::endl;

      for(int i=0; i<matrix.dimensions; i++)
	os << matrix.dim_sizes[i] << " ";

      for(int i=0; i<matrix.planes(); i++)
	os << matrix.get(i);

      return os;
    }

  friend void fwrite(const _DMultiDMatrix<T> &matrix, FILE *fp)
    {
      int rows = matrix.rows(), cols = matrix.cols(), type = sizeof(T);
      
      fwrite(&matrix.dimensions, 1, sizeof(int), fp);
      for(int i=0; i<matrix.dimensions; i++)
	fwrite(&(matrix.dim_sizes[i]), 1, sizeof(int), fp);

      for(int i=0; i<matrix.planes(); i++)
	fwrite(matrix.get(i), fp);
    }

  T max() 
    {
      int plane; DPoint pt;
      return max(plane, pt);
    }

  T max(int &best_plane, DPoint &best_pt)
    {
      best_plane = 0, best_pt = DPoint(0,0);

      T best_val = T(-1e100);

      for(int i=0; i<planes(); i++)
	{
	  DPoint _best_pt;
	  T _best_val = (*this).get(i).max(_best_pt);

	  if(_best_val > best_val)
	    best_pt = _best_pt, best_plane = i, best_val = _best_val;
	}

      return best_val;
    }
    
  _DLazyMultiDMatrix<T>::planes;

 protected:
  int dimensions;
  //int mat_count;
  _DLazyMultiDMatrix<T>::mat_count;

  int *dim_sizes;
  _DMatrix<T> **matrices;
};


typedef _DMultiDMatrix<double> DMultiDMatrix;




//////////////////////////////////////

template<class T>
class _DMatrixArray
{
 public:
  _DMatrixArray() 
    {
      dimensions = 0;
      dim_sizes = 0;
      matrices = 0;
      mat_count=0;
    }

  _DMatrixArray(int _dim_count, int p1)
    {
      assert(_dim_count == 3);
	
      int dims[3];
      dims[0] = p1;
      dims[1] = dims[2] = 0;

      allocate_storage(3, dims);
    }
  
  _DMatrixArray(int _dim_count, int p1, int p2)
    {
      assert(_dim_count == 4);
	
      int dims[4];
      dims[0] = p1;
      dims[1] = p2;
      dims[2] = dims[3] = 0;

      allocate_storage(4, dims);
    }

  _DMatrixArray(int _dim_count, int *_dims)
    {
      assert(_dim_count > 2);

      _dims[_dim_count -1] = _dims[_dim_count - 2] = 0;
      allocate_storage(_dim_count, _dims);
    }
  

  void allocate_storage(int _dim_count, int *_dims)
    {
      dimensions = _dim_count;
      dim_sizes = new int[dimensions];
      for(int i=0; i<dimensions-2; i++)
	dim_sizes[i] = _dims[i];
    
      mat_count = 1;
      for(int i=0; i<dimensions-2; i++)
	mat_count *= _dims[i];
    
      matrices = new _DMatrix<T> *[mat_count];
      for(int i=0; i<mat_count; i++)
	matrices[i] = new _DMatrix<T>;
    }
  
  
  ~_DMatrixArray()
    {
      if(dim_sizes)
	delete[] dim_sizes;
      
      if(matrices)
	{
	  for(int i=0; i<mat_count; i++)
	    delete matrices[i];
	  
	  delete[] matrices;
	}
    }
  
  _DMatrixArray(const _DMatrixArray<T> &other)
    {
      matrices = 0;
      dim_sizes = 0;

      *this = other;
    }


  _DMatrixArray<T> &operator=(const T D)
    {
      for(int i=0; i<mat_count; i++)
	get(i) = D;

      return *this;
    }


  _DMatrixArray<T> &operator=(const _DMatrixArray<T> &other) 
   {

      if(&other == this)
	return *this;

      if(matrices)
	{
	  for(int i=0; i<mat_count; i++)
	    delete matrices[i];

	  delete[] matrices;
	}

      if(dim_sizes)
	delete[] dim_sizes;

      mat_count = other.planes();
      matrices = new _DMatrix<T> *[mat_count];
      dimensions = other.get_dimensions_count();

      for(int i=0; i<mat_count; i++)
	matrices[i] = new _DMatrix<T>(other.get(i));

      dim_sizes = new int[dimensions];
      for(int i=0; i<dimensions; i++)
	dim_sizes[i] = other.get_dims(i);

      return *this;
    }

  int get_dimensions_count() const
    {
      return dimensions;
    }

  _DMatrix<T> &get(int c1) const
    {
      return *(matrices[c1]);
    }

  _DMatrix<T> &get(int c1, int c2) const
    {
      return *(matrices[c1*dim_sizes[1]+c2]);
    }

  //    template<class T>
  //    friend 
  _DMatrixArray<T> pointwise_min(const _DMatrixArray<T> &m1) const; //, const _DMatrixArray<T> &m2);
  //    template<class T>
  //    friend 
  _DMatrixArray<T> pointwise_max(const _DMatrixArray<T> &m1) const; //, const _DMatrixArray<T> &m2);


  int get_dims(int i) const
    {
      return dim_sizes[i];
    }
  //
  template<class T2>
    friend bool same_size(const _DMatrixArray<T2> &m1, const _DMatrixArray<T2> &m2);


  int planes() const { return mat_count; }

  friend std::istream &operator>>(std::istream &is, _DMatrixArray<T> &matrix)
    {
      int dimensions;
      is >> dimensions;

      if(dimensions)
	{
	  int  *dims = new int[ dimensions ];

	  memset (dims, 0, dimensions * sizeof( int ));

	  for(int i=0; i<dimensions; i++)
	    is >> dims[i];
	  
	  matrix = _DMatrixArray<T>(dimensions, dims);
	  for(int i=0; i<matrix.planes(); i++)
	    is >> matrix.get(i);

	  delete [] dims;
	}

      return is;
    }

  friend std::ostream &operator<<(std::ostream &os, const _DMatrixArray<T> &matrix)
    {
      os << matrix.dimensions << std::endl;

      for(int i=0; i<matrix.dimensions; i++)
	os << matrix.dim_sizes[i] << " ";

      for(int i=0; i<matrix.planes(); i++)
	os << matrix.get(i);

      return os;
    }

  friend void fread(_DMatrixArray<T> &matrix, FILE *fp)
    {
      int dimensions;
      fread(&dimensions, 1, sizeof(int), fp);

      if(dimensions)
	{
	  int dims[dimensions];
	  for(int i=0; i<dimensions; i++)
	    fread(dims+i, 1, sizeof(int), fp);

	  matrix = _DMatrixArray<T>(dimensions, dims);

	  for(int i=0; i<matrix.planes(); i++)
	    fread(matrix.get(i), fp);
	}
    }
    

 protected:
  int dimensions;
  int mat_count;

  int *dim_sizes;
  _DMatrix<T> **matrices;
};


typedef _DMatrixArray<double> DMatrixArray;



#endif

