//
//  DLib: A simple image processing library.
//
//  David Crandall, 2003-2005
//  crandall@cs.cornell.edu
//
//  Please do not redistribute this code.
//
//
//
//
#ifndef _DPOINT_H_
#define _DPOINT_H_

#include <assert.h>
#include <iostream>
#include <math.h>

template<class T>
class _DPoint
{
 public:

  _DPoint(const _DPoint<T> &pt) : _row(pt.row()), _col(pt.col()) {}
  _DPoint(T __row, T __col) : _row(__row), _col(__col) {}
  _DPoint() : _row(0), _col(0) {}

  T row() const { return _row; }
  T col() const { return _col; }

  void row(T a) { _row = a; }
  void col(T a) { _col = a; }

  template<class T2> friend _DPoint<T2> operator+(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend _DPoint<T2> operator-(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend _DPoint<T2> operator-(const _DPoint<T2> &p1, double p2);
  template<class T2> friend _DPoint<T2> operator*(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend _DPoint<T2> operator*(const _DPoint<T2> &p1, double factor);
  template<class T2> friend bool operator<(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend bool operator>(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend bool operator<=(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend bool operator>=(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend bool operator==(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend bool operator!=(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend _DPoint<T2> operator/(const _DPoint<T2> &p1, double p2);
  template<class T2> friend _DPoint<T2> operator/(const _DPoint<T2> &p1, const _DPoint<T2> &p2);
  template<class T2> friend _DPoint<T2> sqrt(const _DPoint<T2> &p);
  template<class T2> friend _DPoint<T2> log(const _DPoint<T2> &p);
  template<class T2> friend _DPoint<T2> exp(const _DPoint<T2> &p);

  // just so dplane will compile FIXME
  template<class T2> friend double atan(const _DPoint<T2> &p);

  template<class T2> friend std::ostream &operator<<(std::ostream &os, const _DPoint<T2> &p);
  template<class T2> friend std::istream &operator>>(std::istream &is, _DPoint<T2> &p);

  _DPoint<T> operator-() const
    {
      return _DPoint<T>(-_row, -_col);
    }

  void operator+=(const _DPoint<T> &other) 
    {
      _row += other.row(), _col += other.col();
    }

  void operator/=(const _DPoint<T> &other) 
    {
      _row /= other.row(), _col /= other.col();
    }

  void operator/=(double p) 
    {
      _row /= p, _col /= p;
    }

  void operator*=(double p) 
    {
      _row *= p, _col *= p;
    }

  void operator-=(const _DPoint<T> &other) 
    {
      _row -= other.row(), _col -= other.col();
    }

  void operator*=(const _DPoint<T> &other) 
    {
      _row *= other.row(), _col *= other.col();
    }

  _DPoint<T> swap() const
    {
      return _DPoint<T>(_col, _row);
    }

  T product() const
    {
      return _row * _col;
    }

  T norm_squared() const
    {
      return _row * _row + _col * _col;
    }

  double norm() const
    {
      return sqrt(double(_row * _row + _col * _col));
    }

  _DPoint<T> rotate(float angle) const
    {
      float cos_ = cos(angle), sin_ = sin(angle);

      return _DPoint<T>(T(cos_ * row() + sin_ * col()), T(-sin_ * row() + cos_ * col()));
    }

  friend void fread(_DPoint<T> &pi, FILE *fp)
    {
      fread(&pi._row, 1, sizeof(T), fp);
      fread(&pi._col, 1, sizeof(T), fp);
    }

  friend void fwrite(const _DPoint<T> &pi, FILE *fp)
    {
      fwrite(&pi._row, 1, sizeof(T), fp);
      fwrite(&pi._col, 1, sizeof(T), fp);
    }

  friend _DPoint<T> elementwise_max(const _DPoint<T> &p1, const _DPoint<T> &p2)
    {
      return _DPoint<T>(std::max(p1.row(), p2.row()), std::max(p1.col(), p2.col()));
    }

  friend _DPoint<T> elementwise_min(const _DPoint<T> &p1, const _DPoint<T> &p2)
    {
      return _DPoint<T>(std::min(p1.row(), p2.row()), std::min(p1.col(), p2.col()));
    }

 protected:
  T _row, _col;

};

typedef _DPoint<int> DPoint;

#endif
